/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2012 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.realtime.impl;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * {@link ConnectionStatus}
 *
 * The ConnectionStatus class stores important state information about a
 * document connection. The content is used to be sent to connected clients
 * to update their internal state. The state includes the editor name and
 * id, the current server osn, the number of clients, the current selection
 * of every client etc.
 *
 * @author <a href="mailto:kai.ahrens@open-xchange.com">Kai Ahrens</a>
 * @author <a href="mailto:carsten.driesner@open-xchange.com">Carsten Driesner</a>
 */
public class ConnectionStatus {

    static final String KEY_EDIT_USER = "editUser";

    static final String KEY_EDIT_USERID = "editUserId";

    static final String KEY_HAS_ERRORS = "hasErrors";

    static final String KEY_WRITE_PROTECTED = "writeProtected";

    static final String KEY_LOCKED = "locked";

    static final String KEY_LOCKED_BY_USER = "lockedByUser";

    static final String KEY_FAILSAFESAVEDONE = "failSafeSaveDone";

    static final String KEY_WANTSEDIT_USER = "wantsToEditUser";

    static final String KEY_WANTSEDIT_USERID = "wantsToEditUserId";

    static final String KEY_SERVER_OSN = "serverOSN";

    static final String KEY_ACTIVE_USERS = "activeUsers";

    static final String KEY_LASTMODIFYING_USERID = "lastModifyingUserId";

    // - Implementation --------------------------------------------------------

    /**
     * Initializes a new {@link ConnectionStatus}.
     */
    public ConnectionStatus() {
        super();
    }

    /**
     * Provides the error state.
     *
     * @return The m_hasErrors
     */
    boolean isHasErrors() {
        return m_hasErrors;
    }

    /**
     * Sets the error state.
     *
     * @param hasErrors The m_hasErrors to set
     */
    void setHasErrors(boolean hasErrors) {
        this.m_hasErrors = hasErrors;
    }

    /**
     * Get the lock state
     *
     * @return The current lock state
     */
    boolean isLocked() {
        return m_locked;
    }

    /**
     * Sets the lock state
     *
     * @param locked The new lock state
     */
    void setLockState(boolean locked) {
        this.m_locked = locked;
    }

    /**
     * Get user name who locked the file
     *
     * @return The name of the user who locked the file
     * or empty string if user is unknown or nobody set
     * a lock.
     */
    String getLockedByUser() {
        return m_lockedByUser;
    }

    /**
     * Sets the user who locked the file
     *
     * @param lockedByUser The name of the user
     * or empty string if unknown or nobody set
     * a lock.
     */
    void setLockedByUser(String lockedByUser) {
        this.m_lockedByUser = lockedByUser;
    }

    /**
     * Gets the m_isWriteProtected
     *
     * @return The m_isWriteProtected
     */
    boolean isWriteProtected() {
        return m_isWriteProtected;
    }

    /**
     * Sets the m_isWriteProtected
     *
     * @param isWriteProtected The isWriteProtected to set
     */
    void setWriteProtected(boolean isWriteProtected) {
        this.m_isWriteProtected = isWriteProtected;
    }

    /**
     * Gets the m_currentEditingUserName
     *
     * @return The m_currentEditingUserName
     */
    void setCurrentEditingUser(String currentEditingUserId, String currentEditingUserName) {
        m_currentEditingUserId = currentEditingUserId;
        m_currentEditingUserName = currentEditingUserName;
    }

    /**
     * Gets the m_currentEditingUserName
     *
     * @return The m_currentEditingUserName
     */
    String getCurrentEditingUserId() {
        return m_currentEditingUserId;
    }

    /**
     * Sets the m_currentEditingUserName
     *
     * @param currentEditingUserName The currentEditingUserName to set
     */
    String getCurrentEditingUserName() {
        return m_currentEditingUserName;
    }

    /**
     * Gets the user id who wants to receive the edit rights
     */
    String getWantsEditRightsUserId() {
    	return m_wantsEditRightsUserId;
    }

    /**
     * Gets the user name who wants to receive the edit rights
     * @return
     */
    String getWantsEditRightsUserName() {
    	return m_wantsEditRightsUserName;
    }

    /**
     * Sets the user id who wants to receive the edit rights
     *
     * @param userId
     */
    void setWantsEditRightsUser(String userId, String userName) {
    	m_wantsEditRightsUserId = userId;
    	m_wantsEditRightsUserName = userName;
    }

    /**
     * Provides the number of active clients.
     *
     * @return The number of active clients.
     */
    int getActiveClients() {
        return (null != m_activeUsers) ? m_activeUsers.length() : 0;
    }

    /**
     * Gets the m_operationStateNumber
     *
     * @return The m_operationStateNumber
     */
    int getOperationStateNumber() {
        return this.m_operationStateNumber;
    }

    /**
     * Sets the m_operationStateNumber
     *
     * @param operationStateNumber The operation state number to set
     */
    void setOperationStateNumber(int operationStateNumber) {
        this.m_operationStateNumber = operationStateNumber;
    }

    /**
     * Provides information that a fail safe
     * save operation was done successfully.
     *
     * @return
     *  TRUE if there recently was a successful
     *  fail safe save operation, otherwise FALSE.
     */
    boolean isFailSafeSaveDone() {
        return m_failSafeSaveDone;
    }

    /**
     * Sets the state of a fail safe save
     * operation.
     *
     * @param saveDone The state of the fail safe save
     */
    void setFailSafeSaveDone(boolean saveDone) {
        this.m_failSafeSaveDone = saveDone;
    }

    /**
     * Gets the activeUsers
     *
     * @return The activeUsers
     */
    public JSONArray getActiveUsers() {
        return m_activeUsers;
    }

    /**
     * Sets the activeUsers
     *
     * @param activeUsers The activeUsers to set
     */
    public void setActiveUsers(JSONArray activeUsers) {
        this.m_activeUsers = activeUsers;
    }

    /**
     * Add a user to the activeUsers map
     *
     * @param userId
     *  ID of the user
     * @param userDisplayName
     *  The UI name of the new user
     * @param userData
     *  JSON Object containing user data, e.g. {userName:'Rocky Balboa', selection:{...}, ...}
     * @throws JSONException
     */
    public void addActiveUser(String userId, String userDisplayName, JSONObject userData){
        JSONObject userObject = new JSONObject();
        try {
            userObject.put("userId", userId);
            userObject.put("userDisplayName", userDisplayName);
            userObject.put("userData", userData);
            userObject.put("active", true);
        } catch (JSONException e) {
            LOG.error("Exception catched while adding new active user", e);
        }
        m_activeUsers.put(userObject);
    }

    /**
     * Remove a specified user from the activeUsers map
     *
     * @param userId ID from the user to be removed
     * @throws JSONException
     */
    public boolean removeActiveUser(String userId) throws JSONException {
        for (int i = 0; i < m_activeUsers.length();i++) {
            JSONObject user = m_activeUsers.getJSONObject(i);
            if (user.get("userId").equals(userId)) {
                m_activeUsers.remove(i);
                return true;
            }
        }
        return false;
    }

    /**
     * Update the userdata of a specific user
     *
     * @param userId
     *  ID of the user
     * @param userData
     *  JSON Object containing user data, e.g. {userName:'Rocky Balboa', selection:{...}, ...}
     * @return
     *  Returns true if an update is done.
     */
    public boolean updateUserData(String userId, JSONObject userData){
        // find the user and update his data
        JSONObject user;
        if (null != userId) {
            for (int i = 0; i < m_activeUsers.length();i++) {
                try {
                    user = m_activeUsers.getJSONObject(i);
                    if (userId.equals(user.optString("userId", null))) {
                        user.put("userData", userData);
                        user.put("active", true);
                        return true;
                    }
                } catch (JSONException e) {
                    LOG.error("Exception while setting updated user data", e);
                }
            }
        }

        return false;
    }

    /**
     * Disables an active user so her data is NOT sent to other clients.
     *
     * @param userId
     *  The ID of the user to be disabled.
     * @return
     *  TRUE if the user is disabled, FALSE if the user couldn't be
     *  found in the list of active users.
     */
    public boolean disableActiveUser(String userId) {
        JSONObject user;

        if (null != userId) {
            for (int i = 0; i < m_activeUsers.length();i++) {
                try {
                    user = m_activeUsers.getJSONObject(i);
                    if (userId.equals(user.optString("userId", null))) {
                        user.put("active", false);
                        return true;
                    }
                } catch (JSONException e) {
                    LOG.error("Exception while setting updated user data", e);
                }
            }
        }

        return false;
    }

    /**
     * Determine if the provided user is enabled or not.
     *
     * @param userId
     *  The ID of the user to find out the enabled state.
     *
     * @return
     *  TRUE if the user is enabled, FALSE if the user couldn't be
     *  found in the list or is not enabled.
     */
    public boolean isActiveUserEnabled(String userId) {
        JSONObject user;

        if (null != userId) {
            try {
                for (int i = 0; i < m_activeUsers.length();i++) {
                    user = m_activeUsers.getJSONObject(i);
                    if (userId.equals(user.optString("userId", null))) {
                        return user.optBoolean("active", false);
                    }
                }
            } catch (JSONException e) {
                LOG.error("Exception while setting updated user data", e);
            }
        }

        return false;
    }

    /**
     * Enable an active user again so her data is sent to other clients.
     *
     * @param userId
     *  The ID of the user to be enabled again.
     * @return
     *  TRUE if the user is active again, FALSE if the user couldn't be
     *  found in the list of active users.
     *
     * @throws JSONException
     */
    public boolean enableActiveUser(String userId) {
        JSONObject user;
        for (int i = 0; i < m_activeUsers.length();i++) {
            try {
                user = m_activeUsers.getJSONObject(i);
                if (user.get("userId").equals(userId)) {
                    user.put("active", true);
                    return true;
                }
            } catch (JSONException e) {
                LOG.error("Exception while setting updated user data", e);
            }
        }
        return false;
    }

    /**
     * Provides the userid that made the latest change to the
     * document.
     *
     * @return
     *  The userid or an empty string if no modification has been
     *  done yet.
     */
    public String getLastModifyingUserId() {
        return m_lastModifyingUserId;
    }

    /**
     * Set the userid of the user who made the latest change to
     * the document.
     *
     * @param userId
     *  The userid as a string. Must not be null or empty.
     */
    public void setLastModifyingUserId(String userId) {
        if (null != userId) {
            m_lastModifyingUserId = userId;
        }
    }

    /**
     * Creates a shallow clone of this connection status
     * object.
     *
     * @return
     *  A shallow clone.
     */
    @Override
    public Object clone() {
        ConnectionStatus clone = new ConnectionStatus();
        clone.setCurrentEditingUser(m_currentEditingUserId, m_currentEditingUserName);
        clone.setFailSafeSaveDone(m_failSafeSaveDone);
        clone.setHasErrors(m_hasErrors);
        clone.setLockedByUser(m_lockedByUser);
        clone.setLockState(m_locked);
        clone.setOperationStateNumber(m_operationStateNumber);
        clone.setWantsEditRightsUser(m_wantsEditRightsUserId, m_wantsEditRightsUserName);
        clone.setWriteProtected(m_isWriteProtected);
        clone.setActiveUsers(m_activeUsers);
        clone.setLastModifyingUserId(m_lastModifyingUserId);
        return clone;
    }

    /**
     * Creates the JSON representation of the current connection status
     * object. The JSON object should only provide properties that are
     * part of the client/server communication. Internal properties should
     * be omitted.
     *
     * @return
     *  A JSON object that contains the properties of the current connection
     *  status. The JSON object can be empty in case of an error.
     */
    public JSONObject toJSON() {
        final JSONObject ret = new JSONObject();

        try {
            ret.put(KEY_HAS_ERRORS, m_hasErrors);
            ret.put(KEY_WRITE_PROTECTED, m_isWriteProtected);
            ret.put(KEY_EDIT_USERID, m_currentEditingUserId);
            ret.put(KEY_EDIT_USER, m_currentEditingUserName);
            ret.put(KEY_LOCKED, m_locked);
            ret.put(KEY_LOCKED_BY_USER, m_lockedByUser);
            ret.put(KEY_FAILSAFESAVEDONE, m_failSafeSaveDone);
            ret.put(KEY_WANTSEDIT_USER, m_wantsEditRightsUserName);
            ret.put(KEY_WANTSEDIT_USERID, m_wantsEditRightsUserId);
            ret.put(KEY_SERVER_OSN, m_operationStateNumber);
            // Filter out active users that are temporary disabled
            // due to the fact that their client didn't respond
            // in time to edit rights changes.
            JSONArray filteredActiveUsers = new JSONArray();
            for (int i = 0; i < m_activeUsers.length(); i++) {
                JSONObject obj = m_activeUsers.getJSONObject(i);
                if (null != obj && obj.optBoolean("active", false)) {
                    filteredActiveUsers.put(obj);
                }
            }
            ret.put(KEY_ACTIVE_USERS, filteredActiveUsers);
        } catch (JSONException e) {
            LOG.error("Exception while creating JSON object of the current connection status", e);
        }

        return ret;
    }

    /**
     * Creates a new connection status object using the provided JSONObject
     * to set the state.
     *
     * @param jsonData
     *  The JSON object containing properties that are used to set the
     *  state of the new connection status object.
     *
     * @return
     *  The new connection status object.
     */
    static public ConnectionStatus createFrom(JSONObject jsonData) {
        ConnectionStatus queueStatus = new ConnectionStatus();
        boolean found = false;

        try {
            if (jsonData.has(KEY_HAS_ERRORS)) {
                queueStatus.m_hasErrors = jsonData.getBoolean(KEY_HAS_ERRORS);
                found = true;
            }

            if (jsonData.has(KEY_WRITE_PROTECTED)) {
                queueStatus.m_isWriteProtected = jsonData.getBoolean(KEY_WRITE_PROTECTED);
                found = true;
            }

            if (jsonData.has(KEY_EDIT_USERID)) {
                queueStatus.m_currentEditingUserId = jsonData.getString(KEY_EDIT_USERID);
                found = true;
            }

            if (jsonData.has(KEY_EDIT_USER)) {
                queueStatus.m_currentEditingUserName = jsonData.getString(KEY_EDIT_USER);
                found = true;
            }

            if (jsonData.has(KEY_LOCKED)) {
                queueStatus.m_locked = jsonData.getBoolean(KEY_LOCKED);
                found = true;
            }

            if (jsonData.has(KEY_LOCKED_BY_USER)) {
                queueStatus.m_lockedByUser = jsonData.getString(KEY_LOCKED_BY_USER);
                found = true;
            }

            if (!jsonData.has(KEY_FAILSAFESAVEDONE)) {
                queueStatus.m_failSafeSaveDone = jsonData.getBoolean(KEY_FAILSAFESAVEDONE);
                found = true;
            }

            if (!jsonData.has(KEY_WANTSEDIT_USER)) {
                queueStatus.m_wantsEditRightsUserName = jsonData.getString(KEY_WANTSEDIT_USER);
                found = true;
            }

            if (!jsonData.has(KEY_WANTSEDIT_USERID)) {
                queueStatus.m_wantsEditRightsUserId = jsonData.getString(KEY_WANTSEDIT_USERID);
                found = true;
            }

            if (!jsonData.has(KEY_ACTIVE_USERS)) {
                queueStatus.m_activeUsers = jsonData.getJSONArray(KEY_ACTIVE_USERS);
                found = true;
            }
        } catch (JSONException e) {
            LOG.error("Exception while creating new connection status object", e);
        }

        return (found ? queueStatus : null);
    }

    // - Members ---------------------------------------------------------------

    static protected final org.apache.commons.logging.Log LOG = com.openexchange.log.LogFactory.getLog(ConnectionStatus.class);
    private boolean m_hasErrors = false;
    private boolean m_isWriteProtected = false;
    private boolean m_locked = false;
    private String m_currentEditingUserName = "";
    private String m_currentEditingUserId = "";
    private String m_wantsEditRightsUserId = "";
    private String m_wantsEditRightsUserName = "";
    private String m_lastModifyingUserId = "";
    private int m_operationStateNumber = 0;
    private String m_lockedByUser = "";
    private boolean m_failSafeSaveDone = false;
    private JSONArray m_activeUsers = new JSONArray() ;
}
